# copyright (c) 2020 PaddlePaddle Authors. All Rights Reserve.
#
# Modifications copyright (c) 2021 DocYard Authors. All Rights Reserve.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#    http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import os
from tqdm import tqdm
import cv2
import copy
import numpy as np
import pandas as pd

from PIL import Image
import hydra
from omegaconf import OmegaConf
from hydra.experimental import compose, initialize_config_dir

import sys
__dir__ = os.path.dirname(__file__)
sys.path.append(os.path.join(__dir__, '../..'))

import ucr.inference.infer_rec as infer_rec
import ucr.inference.infer_det as infer_det
import ucr.inference.infer_cls as infer_cls
from ucr.utils.utility import get_image_file_list, check_and_read_gif, download_with_progressbar, merge_text_boxes, sorted_boxes
from ucr.utils.annotation import draw_ocr_box_txt, draw_text_det_res

class TextSystem(object):
    def __init__(self, config):
        self.text_detector = infer_det.TextDetector(config['Detection'])
        self.text_recognizer = infer_rec.TextRecognizer(config['Recognition'])
        self.text_classifier = infer_cls.TextClassifier(config['Classification'])
        
        self.drop_score = config['drop_score']            
        self.merge_boxes = config['merge_boxes']
        self.output_format = config['output_format']
        self.verbose = config['verbose']
        self.font_path = config['Recognition']['font_path']
        
        if self.merge_boxes:
            self.merge_slope_thresh = config['merge_slope_thresh']
            self.merge_ycenter_thresh = config['merge_ycenter_thresh']
            self.merge_height_thresh = config['merge_height_thresh']
            self.merge_width_thresh = config['merge_width_thresh']
            self.merge_add_margin = config['merge_add_margin']

    def get_rotate_crop_image(self, img, points):
        '''
        img_height, img_width = img.shape[0:2]
        left = int(np.min(points[:, 0]))
        right = int(np.max(points[:, 0]))
        top = int(np.min(points[:, 1]))
        bottom = int(np.max(points[:, 1]))
        img_crop = img[top:bottom, left:right, :].copy()
        points[:, 0] = points[:, 0] - left
        points[:, 1] = points[:, 1] - top
        '''
        img_crop_width = int(
            max(
                np.linalg.norm(points[0] - points[1]),
                np.linalg.norm(points[2] - points[3])))
        img_crop_height = int(
            max(
                np.linalg.norm(points[0] - points[3]),
                np.linalg.norm(points[1] - points[2])))
        pts_std = np.float32([[0, 0], [img_crop_width, 0],
                              [img_crop_width, img_crop_height],
                              [0, img_crop_height]])
        M = cv2.getPerspectiveTransform(points, pts_std)
        dst_img = cv2.warpPerspective(
            img,
            M, (img_crop_width, img_crop_height),
            borderMode=cv2.BORDER_REPLICATE,
            flags=cv2.INTER_CUBIC)
        dst_img_height, dst_img_width = dst_img.shape[0:2]
        if dst_img_height * 1.0 / dst_img_width >= 1.5:
            dst_img = np.rot90(dst_img)
        return dst_img

    def det_cls_rec(self, img, cls):  
        ori_im = img.copy()
        dt_boxes, elapse = self.text_detector(img)
        if self.verbose:
            print("\n------------------------dt_boxes num : {},\telapse : {:.3f}------------------------".format(
                len(dt_boxes), elapse))
        if dt_boxes is None:
            return None, None
        img_crop_list = []

        dt_boxes = sorted_boxes(dt_boxes)

        for bno in range(len(dt_boxes)):
            tmp_box = copy.deepcopy(dt_boxes[bno])
            img_crop = self.get_rotate_crop_image(ori_im, tmp_box)
            img_crop_list.append(img_crop)
        
        if cls:  
            img_crop_list, _, elapse = self.text_classifier(
                img_crop_list)
            if self.verbose:
                print("------------------------cls num  : {},\telapse : {:.3f}------------------------".format(
                    len(img_crop_list), elapse))

        rec_res, elapse = self.text_recognizer(img_crop_list)
        if self.verbose:
            print("------------------------rec_res num  : {},\telapse : {:.3f}------------------------\n".format(
            len(rec_res), elapse))

        filter_boxes, filter_rec_res = [], []
        
        if self.drop_score!=0.:
            for box, rec_reuslt in zip(dt_boxes, rec_res):
                _, score = rec_reuslt
                if score >= self.drop_score:
                    filter_boxes.append(box)
                    filter_rec_res.append(rec_reuslt) 
        else:
            filter_boxes = dt_boxes
            filter_rec_res = rec_res
                
        if self.merge_boxes:
            free_box, free_text, merged_box, merged_text = merge_text_boxes(
                filter_boxes, filter_rec_res,
                slope_thresh = self.merge_slope_thresh,
                ycenter_thresh = self.merge_ycenter_thresh,
                height_thresh = self.merge_height_thresh,
                width_thresh = self.merge_width_thresh,
                add_margin = self.merge_add_margin)
            dt_boxes = free_box + merged_box
            rec_res = free_text + merged_text
            
        return dt_boxes, rec_res
    
    def perform_ocr(self, img, key, output, rec, cls):
        if rec:
            dt_boxes, rec_res = self.det_cls_rec(img, cls)
            if self.merge_boxes:
                rec_res = rec_res[0:1]
            
            if self.output_format=='ppocr':
                value = [[box.tolist(), res] for box, res in zip(dt_boxes, rec_res)]
            elif self.output_format=='df':
                info_list = [[int(box[0][0]), int(box[0][1]), int(box[2][0]), int(box[2][1]), res[0]] for box, res in zip(dt_boxes, rec_res)]
                value = pd.DataFrame(info_list, columns=['startX', 'startY', 'endX', 'endY', 'Text'])
                
            if self.verbose:
                from tabulate import tabulate
                headers = ["OCR Result", "Score"]
                if self.merge_boxes:
                    print(tabulate(rec_res[0:1], headers, tablefmt="fancy_grid"))
                else:
                    print(tabulate(rec_res, headers, tablefmt="fancy_grid"))
                    
            if output:
                image = Image.fromarray(cv2.cvtColor(img, cv2.COLOR_BGR2RGB))
                boxes = dt_boxes
                txts = [rec_res[i][0] for i in range(len(rec_res))]
                scores = [rec_res[i][1] for i in range(len(rec_res))]

                draw_img = draw_ocr_box_txt(
                    image,
                    boxes,
                    txts,
                    scores,
                    drop_score=self.drop_score,
                    font_path=self.font_path)
                
                if not os.path.exists(output):
                    os.makedirs(output)
                img_path = os.path.join(output,
                                        "ocr_{}".format(key))
                cv2.imwrite(
                    img_path, draw_img[:, :, ::-1])
                print("Saving OCR ouput in {}".format(
                    img_path))
            
            return value
            
        else:
            dt_boxes, elapse = self.text_detector(img)
            if self.verbose:
                print("\n------------------------dt_boxes num : {},\telapse : {:.3f}------------------------".format(
                len(dt_boxes), elapse))
            if dt_boxes is not None:
                if self.output_format=='ppocr':
                    value = [box.tolist() for box in dt_boxes]
                    
                elif self.output_format=='df':
                    info_list = [[int(box[0][0]), int(box[0][1]), int(box[2][0]), int(box[2][1])] for box in dt_boxes]
                    value = pd.DataFrame(info_list, columns=['startX', 'startY', 'endX', 'endY'])
                
                if output:
                    src_im = draw_text_det_res(dt_boxes, img)
                    
                    if not os.path.exists(output):
                        os.makedirs(output)
                    img_path = os.path.join(output,
                                            "det_{}".format(key))
                    cv2.imwrite(img_path, src_im)
                    print("Detection output image is saved in {}".format(img_path))
            
            return value
        
    
    def __call__(self, input=None, i=None, output=None, o=None, det=True, rec=True, cls=False):
        
        if input is not None:
            input = input
        elif i is not None:
            input = i
        elif input is None and i is None:
            print('ERROR: Input is mandatory: can be either ndarray (or list of ndarray), img_filepath, img_folderpath or img_webpath!')
            sys.exit(0)
        
        assert isinstance(input, (np.ndarray, list, str))
        
        rec_dict = {}
        out_dict={} 
        
        if output is not None:
            output = output
        elif o is not None:
            output = o
            
        if output:
            print("Saving prediction results in '{}' folder\n".format(output))  
            if not os.path.exists(output):
                os.makedirs(output)   
        
        is_imgpath = False
        if isinstance(input, str):
            # download net image
            if input.startswith('http') and input.endswith('.jpg'):
                download_with_progressbar(input, 'downloaded.jpg')
                input = 'downloaded.jpg'
                
            input = get_image_file_list(input)
            is_imgpath = True
            
        elif isinstance(input, np.ndarray):
            input = [input]
        
        else:
            if isinstance(input[0], str):
                is_imgpath = True
                
        if len(input) == 0:
            print('NO images found in {}. Please check input location!'.format(input))
            sys.exit(0) 
        
        print(f'Performing Inference on {len(input)} files')
        i = 0
        for image in tqdm(input, colour='green', desc='OCR', unit='image'):
            if is_imgpath:
                key = image
                img, flag = check_and_read_gif(image)
                if not flag:
                    img = cv2.imread(image)
                if img is None:
                    print("ERROR in loading image:{}".format(image))
                    continue
            else:
                key = str(i) + '.jpg'
                i +=1
                if len(image.shape) == 2:
                    img = cv2.cvtColor(image, cv2.COLOR_GRAY2BGR) 
                elif len(image.shape) == 3:
                    img = cv2.cvtColor(image, cv2.COLOR_RGB2BGR) 
                elif len(image.shape) == 4: # png file
                    img = cv2.cvtColor(image[:3], cv2.COLOR_RGB2BGR) 
                else:
                    print("ERROR: Input array has {} channels. Expected '2', '3' or '4'.".format(len(image.shape)))
                    continue  
            if det:
                out_dict[key] = self.perform_ocr(img, key, output, rec, cls)
            else:
                rec_dict[key] = img
                    
        if rec_dict:
            img = list(rec_dict.values())
            if cls:
                img, cls_res, _ = self.text_classifier(img)
                if not rec:
                    if self.verbose:
                        cls_list = [[k,v[0], v[1]] for k,v in zip(list(rec_dict.keys()), cls_res)]
                        print("Classification Result:\n")
                        from tabulate import tabulate
                        headers = ["File Name", "Classification Result", "Score"]
                        print(tabulate(cls_list, headers, tablefmt="fancy_grid"))
                        
                    if output:
                        output_names = [f'{k}_{v[0]}.jpg' for k,v in zip(list(rec_dict.keys()), cls_res)]    
                        for bno in range(len(img)):
                            cv2.imwrite(os.path.join(output,output_names[bno]), img[bno])
                        
                    cls_output = {k:[v] for k,v in zip(list(rec_dict.keys()), cls_res)}       
                    return cls_output               
            if rec:    
                rec_res, _ = self.text_recognizer(img)
                if self.verbose:
                    rec_list = [[k,v[0], v[1]] for k,v in zip(list(rec_dict.keys()), rec_res)]
                    print("Recognition Result:\n")
                    from tabulate import tabulate
                    headers = ["File Name", "Recognition Result", "Score"]
                    print(tabulate(rec_list, headers, tablefmt="fancy_grid"))
                                            
                if output:
                    output_names = [f'{k}_{v[0]}.jpg' for k,v in zip(list(rec_dict.keys()), rec_res)]    
                    for bno in range(len(img)):
                        cv2.imwrite(os.path.join(output,output_names[bno]), img[bno])
                        
                rec_output = {k:[v] for k,v in zip(list(rec_dict.keys()), rec_res)} 
                return rec_output
        else:
            return out_dict


def main():        
    cfg_dir = hydra.utils.to_absolute_path('conf')

    with initialize_config_dir(config_dir=cfg_dir, job_name="infer_det"):
        cfg = compose(config_name="infer_det")    
        # print("Detection config:\n{}\n".format(cfg.pretty()))   
        config_det = OmegaConf.to_container(cfg)
        
    with initialize_config_dir(config_dir=cfg_dir, job_name="infer_rec"):
        cfg = compose(config_name="infer_rec")      
        # print("Recognition config:\n{}\n".format(cfg.pretty()))   
        config_rec = OmegaConf.to_container(cfg)
        
    with initialize_config_dir(config_dir=cfg_dir, job_name="infer_cls"):
        cfg = compose(config_name="infer_cls")       
        # print("Classification config:\n{}\n".format(cfg.pretty()))  
        config_cls = OmegaConf.to_container(cfg)
    
    with initialize_config_dir(config_dir=cfg_dir, job_name="infer_system"):
        cfg = compose(config_name="infer_system")        
        # print("Final config:\n{}\n".format(cfg.pretty())) 
        config = OmegaConf.to_container(cfg)
        
    config['Detection'] = config_det
    config['Recognition'] = config_rec
    config['Classification'] = config_cls
    
    text_sys = TextSystem(config)
    
    input = hydra.utils.to_absolute_path(config['input'])
    _ = text_sys(input)
    

if __name__ == "__main__":
    main()



# copyright (c) 2020 PaddlePaddle Authors. All Rights Reserve.
#
# Modifications copyright (c) 2021 DocYard Authors. All Rights Reserve.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#    http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

from __future__ import absolute_import
from __future__ import division
from __future__ import print_function

import numpy as np
import hydra
from hydra.core.global_hydra import GlobalHydra
from omegaconf import OmegaConf
import os
import sys
import logging

log = logging.getLogger(__name__)
__dir__ = os.path.dirname(os.path.abspath(__file__))
sys.path.append(__dir__)
sys.path.append(os.path.abspath(os.path.join(__dir__, '../..')))

from tqdm import tqdm
import cv2
import time
import traceback
import torch

import ucr.utils.annotation as utility
from ucr.core.architecture import build_architecture
from ucr.core.postprocess import build_postprocess
from ucr.core.preprocess import build_preprocess, preprocess
from ucr.utils.utility import get_image_file_list, check_and_read_gif
from ucr.core.preprocess.label_ops import BaseRecLabelEncode


class TextRecognizer(object):
    def __init__(self, config):
        self.device = config['device']
        
        for op in config['Preprocess']:
            op_name = list(op)[0]
            if op_name == 'RecResizeImg':
                image_shape = op[op_name]['image_shape'] # TODO:add try except here
                
        self.rec_image_shape = image_shape
        
        self.character_type = config['lang']
        self.rec_batch_num = config['batch_size']
        self.rec_algorithm = config['Architecture']['algorithm']
        self.rec_whitelist = config['whitelist']
        self.rec_blacklist = config['blacklist']
        
        # Todo: Implement whitelist and blacklist in postprocessing step.
        
        char_dict_location = config['char_dict_location']
        self.char_ops = BaseRecLabelEncode(config['max_text_length'], char_dict_location, config['lang'], config['use_space_char'])
            
        global_keys = ['lang', 'use_space_char', 'max_text_length']
        global_cfg = {key: value for key, value in config.items() if key in global_keys}
        global_cfg['char_dict_location'] = char_dict_location
        
            
        self.preprocess_op = build_preprocess(config['Preprocess'])
        self.postprocess_op = build_postprocess(config['Postprocess'], global_cfg)
        
        # build model
        if hasattr(self.postprocess_op, 'character'):
            config['Architecture']["Head"]['out_channels'] = len(
                getattr(self.postprocess_op, 'character'))
        self.predictor = build_architecture(config['Architecture'])

        if self.device == 'cuda':
            self.device = torch.device('cuda') if torch.cuda.is_available() else torch.device('cpu')
        elif self.device == 'cpu':
            self.device = torch.device('cpu')
        else:
            log.error("wrong device selected! Choose eiter 'cuda' or 'cpu'")
            sys.exit(0)
        self.predictor.load_state_dict(torch.load(config['model_location'], map_location=self.device))
        self.predictor.eval()

    def __call__(self, img_list):
        img_num = len(img_list)
        # Calculate the aspect ratio of all text bars
        width_list = []
        for img in img_list:
            width_list.append(img.shape[1] / float(img.shape[0]))
        # Sorting can speed up the recognition process
        indices = np.argsort(np.array(width_list))

        # rec_res = []
        rec_res = [['', 0.0]] * img_num
        batch_num = self.rec_batch_num
        elapse = 0
        for beg_img_no in range(0, img_num, batch_num):
            end_img_no = min(img_num, beg_img_no + batch_num)
            norm_img_batch = []
            for ino in range(beg_img_no, end_img_no):
                
                image = {'image': img_list[indices[ino]]}
                norm_img = preprocess(image, self.preprocess_op)
                
                if self.rec_algorithm != "SRN":
                    norm_img = norm_img['image'][np.newaxis, :]
                    norm_img_batch.append(norm_img)
                else:
                    encoder_word_pos_list = []
                    gsrm_word_pos_list = []
                    gsrm_slf_attn_bias1_list = []
                    gsrm_slf_attn_bias2_list = []
                    encoder_word_pos_list.append(norm_img[1])
                    gsrm_word_pos_list.append(norm_img[2])
                    gsrm_slf_attn_bias1_list.append(norm_img[3])
                    gsrm_slf_attn_bias2_list.append(norm_img[4])
                    norm_img_batch.append(norm_img[0])
            norm_img_batch = np.concatenate(norm_img_batch)
            norm_img_batch = norm_img_batch.copy()

            if self.rec_algorithm == "SRN":
                starttime = time.time()
                encoder_word_pos_list = np.concatenate(encoder_word_pos_list)
                gsrm_word_pos_list = np.concatenate(gsrm_word_pos_list)
                gsrm_slf_attn_bias1_list = np.concatenate(
                    gsrm_slf_attn_bias1_list)
                gsrm_slf_attn_bias2_list = np.concatenate(
                    gsrm_slf_attn_bias2_list)
                
                inputs = [
                    torch.as_tensor(norm_img_batch).to(self.device),
                    torch.as_tensor(encoder_word_pos_list).to(self.device),
                    torch.as_tensor(gsrm_word_pos_list).to(self.device),
                    torch.as_tensor(gsrm_slf_attn_bias1_list).to(self.device),
                    torch.as_tensor(gsrm_slf_attn_bias2_list).to(self.device),
                ]
                self.predictor.to(self.device)
                
                output_tensors = self.predictor(inputs[0], inputs[1:])
                outputs = []
                for output_tensor in output_tensors.values():
                    output = output_tensor.cpu().data.numpy()
                    outputs.append(output)
                preds = {"predict": outputs[3]}
            else:
                starttime = time.time()
                imgs = torch.as_tensor(norm_img_batch)
                input_tensors = imgs.to(self.device)
                self.predictor.to(self.device)
                
                output_tensors = self.predictor(input_tensors)

                preds = output_tensors.cpu().data.numpy()
                # If both blacklist and whitelist are provided, whitelist is only used
                if not self.rec_whitelist and self.rec_blacklist:
                    self.mod_chars = np.arange(preds.shape[-1])
                    black_list = self.char_ops.encode(self.rec_blacklist)
                    black_list = np.array(black_list) + 1
                    self.mod_chars = np.setdiff1d(self.mod_chars, black_list)
                elif self.rec_whitelist:
                    white_list = self.char_ops.encode(self.rec_whitelist)
                    self.mod_chars = np.append(white_list, [-1]) + 1
                elif not self.rec_whitelist and not self.rec_blacklist:
                    self.mod_chars = []
                    
                if len(self.mod_chars)!=0:
                    mod_onehot = np.zeros(preds.shape[-1])
                    mod_onehot[self.mod_chars] = 1
                    preds = np.multiply(preds, mod_onehot) #* Implemented blacklist and whitelist here!

            rec_result = self.postprocess_op(preds)
            for rno in range(len(rec_result)):
                rec_res[indices[beg_img_no + rno]] = rec_result[rno]
            elapse += time.time() - starttime
        return rec_res, elapse


def main(cfg):
    log.debug("Recognition config:\n{}\n".format(cfg.pretty()))  
    config = OmegaConf.to_container(cfg)
    GlobalHydra.instance().clear()
    
    input = hydra.utils.to_absolute_path(config['input'])
    model_location = hydra.utils.to_absolute_path(config['model_location'])
    config['model_location'] = model_location
    char_dict_location = hydra.utils.to_absolute_path(config['char_dict_location'])
    config['char_dict_location'] = char_dict_location
    
    image_file_list = get_image_file_list(input)
    text_recognizer = TextRecognizer(config)
    valid_image_file_list = []
    img_list = []
    for image_file in tqdm(image_file_list, colour='green', desc='Recognition', unit = 'image'):
        img, flag = check_and_read_gif(image_file)
        if not flag:
            img = cv2.imread(image_file)
        if img is None:
            log.warning("error in loading image:{}".format(image_file))
            continue
        valid_image_file_list.append(image_file)
        img_list.append(img)
    try:
        rec_res, predict_time = text_recognizer(img_list)
    except:
        log.error(traceback.format_exc())
        log.error(
            "ERROR!!!! \n"
            "Please read the FAQ：https://github.com/PaddlePaddle/PaddleOCR#faq \n"
            "If your model has tps module:  "
            "TPS does not support variable shape.\n"
            "Please set --rec_image_shape='3,32,100' and --rec_char_type='en' ")
        exit()
    for ino in range(len(img_list)):
        log.info("[{}/{}] Prediction ['RESULT', SCORE] of {}:\t{}".format(ino+1,len(img_list),valid_image_file_list[ino],
                                               rec_res[ino]))
    log.info("\nTotal Prediction time for {} images:\t{:.5f} s".format(
        len(img_list), predict_time))


if __name__ == "__main__":
    import argparse
    parser = argparse.ArgumentParser()
    parser.add_argument("--config_path", type=str, default='../../conf')
    parser.add_argument("--config_name", type=str, default='infer_rec')
    
    args = parser.parse_args()
    
    main_wrapper = hydra.main(config_path=args.config_path, config_name=args.config_name)
    main_wrapper(main)()




# Copyright (c) 2020 PaddlePaddle Authors. All Rights Reserved.
#
# Modifications copyright (c) 2021 DocYard Authors. All Rights Reserve.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

from __future__ import absolute_import
from __future__ import division
from __future__ import print_function

import hydra
from hydra.core.global_hydra import GlobalHydra
from omegaconf import OmegaConf
import numpy as np
import traceback
import os
import sys
import logging
from tqdm import tqdm

log = logging.getLogger(__name__)

__dir__ = os.path.dirname(os.path.abspath(__file__))
sys.path.append(__dir__)
sys.path.append(os.path.abspath(os.path.join(__dir__, '../..')))

import torch
import cv2
import copy
import time

from ucr.core.preprocess import build_preprocess, preprocess
from ucr.core.postprocess import build_postprocess
from ucr.core.architecture import build_architecture
from ucr.utils.utility import get_image_file_list, check_and_read_gif


class TextClassifier(object):
    def __init__(self, config):
        self.device = config['device']
        
        for op in config['Preprocess']:
            op_name = list(op)[0]
            if op_name == 'ClsResizeImg':
                image_shape = op[op_name]['image_shape'] # TODO:add try except here
                
        self.cls_image_shape = image_shape
        
        self.batch_size = config['batch_size']
        # * Note: implement cls_image_shape in config['Postprocess'] Phase 2
        self.threshold = config['threshold']
        
        label_list = {'label_list': config['label_list']}
        
        self.preprocess_op = build_preprocess(config['Preprocess'])
        self.postprocess_op = build_postprocess(config['Postprocess'], label_list)
        self.predictor = build_architecture(config['Architecture'])

        if self.device == 'cuda':
            self.device = torch.device('cuda') if torch.cuda.is_available() else torch.device('cpu')
        elif self.device == 'cpu':
            self.device = torch.device('cpu')
        else:
            log.error("wrong device selected! Choose eiter 'cuda' or 'cpu'")
            sys.exit(0)
        self.predictor.load_state_dict(torch.load(config['model_location'], map_location=self.device))
        self.predictor.eval()

    def __call__(self, img_list):
        img_list = copy.deepcopy(img_list)
        img_num = len(img_list)
        # Calculate the aspect ratio of all text bars
        width_list = []
        for img in img_list:
            width_list.append(img.shape[1] / float(img.shape[0]))
        # Sorting can speed up the cls process
        indices = np.argsort(np.array(width_list))

        cls_res = [['', 0.0]] * img_num
        batch_num = self.batch_size
        elapse = 0
        for beg_img_no in range(0, img_num, batch_num):
            end_img_no = min(img_num, beg_img_no + batch_num)
            norm_img_batch = []
            for ino in range(beg_img_no, end_img_no):
                image = {'image': img_list[indices[ino]]}
                norm_img = preprocess(image, self.preprocess_op)['image']
                norm_img = norm_img[np.newaxis, :]
                norm_img_batch.append(norm_img)
            norm_img_batch = np.concatenate(norm_img_batch)
            starttime = time.time()

            # self.input_tensor.copy_from_cpu(norm_img_batch)
            input_tensors = torch.as_tensor(norm_img_batch)
            input_tensors = input_tensors.to(self.device)            
            self.predictor.to(self.device)
            output_tensors = self.predictor(input_tensors)
            prob_out = output_tensors.cpu().data.numpy()
            
            cls_result = self.postprocess_op(prob_out)
            elapse += time.time() - starttime
            for rno in range(len(cls_result)):
                label, score = cls_result[rno]
                cls_res[indices[beg_img_no + rno]] = [label, score]
                if '180' in label and score > self.threshold:
                    img_list[indices[beg_img_no + rno]] = cv2.rotate(
                        img_list[indices[beg_img_no + rno]], 1)
        return img_list, cls_res, elapse


def main(cfg):
    log.debug("Classification config:\n{}\n".format(cfg.pretty()))  
    config = OmegaConf.to_container(cfg)
    GlobalHydra.instance().clear()
    
    input = hydra.utils.to_absolute_path(config['input'])
    model_location = hydra.utils.to_absolute_path(config['model_location'])
    config['model_location'] = model_location
    
    image_file_list = get_image_file_list(input)
    text_classifier = TextClassifier(config)
    valid_image_file_list = []
    img_list = []
    for image_file in tqdm(image_file_list, colour='green', desc='Classification', unit = 'image'):
        img, flag = check_and_read_gif(image_file)
        if not flag:
            img = cv2.imread(image_file)
        if img is None:
            log.warning("error in loading image:{}".format(image_file))
            continue
        valid_image_file_list.append(os.path.basename(image_file))
        img_list.append(img)
    try:
        img_list, cls_res, predict_time = text_classifier(img_list)
    except:
        log.error(traceback.format_exc())
        log.error(
            "ERROR!!!! \n"
            "Please read the FAQ：https://github.com/PaddlePaddle/PaddleOCR#faq \n"
            "If your model has tps module:  "
            "TPS does not support variable shape.\n"
            "Please set --rec_image_shape='3,32,100' and --rec_char_type='en' ")
        exit()
    for ino in range(len(img_list)):
        log.info("[{}/{}] Prediction ['RESULT', SCORE] of {}:\t{}".format(ino+1,len(img_list),valid_image_file_list[ino],
                                               cls_res[ino]))
    log.info("\nTotal Prediction time for {} images:\t{:.5f} s".format(
        len(img_list), predict_time))


if __name__ == "__main__":
    import argparse
    parser = argparse.ArgumentParser()
    parser.add_argument("--config_path", type=str, default='../../conf')
    parser.add_argument("--config_name", type=str, default='infer_cls')
    
    args = parser.parse_args()
    
    main_wrapper = hydra.main(config_path=args.config_path, config_name=args.config_name)
    main_wrapper(main)()




# Copyright (c) 2020 PaddlePaddle Authors. All Rights Reserved.
#
# Modifications copyright (c) 2021 DocYard Authors. All Rights Reserve.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import os

from tabulate import tabulate
from omegaconf import OmegaConf
from hydra.experimental import compose, initialize_config_dir

from ucr.inference import infer_system
from ucr.utils.utility import maybe_download

__all__ = ['UCR']

__dir__ = os.path.dirname(os.path.abspath(__file__))
SUPPORT_DET_MODEL = ['DB', 'CRAFT']
SUPPORT_REC_MODEL = ['CRNN']
SUPPORT_CLS_MODEL = ['CLS']
VERSION = 1.0
SUPPORT_MODEL_TYPE = ['torch_mobile', 'torch_server', 'onnx_mobile', 'onnx_server']
BASE_DIR = os.path.expanduser("~/.ucr/")


def parse_args(mMain=True, add_help=True):
    import argparse
    
    def str2bool(v):
        return v.lower() in ("true", "t", "1")

    if mMain:
        parser = argparse.ArgumentParser()
        
        # params for prediction system
        parser.add_argument("--hydra_conf_location", type=str, default=None)
        parser.add_argument("--force_download", type=str2bool, default=False)
        parser.add_argument("-i", "--input", type=str, required=True)
        parser.add_argument("-o", "--output", type=str, default=None)
        parser.add_argument("-d", "--device", type=str, default='cpu')
        parser.add_argument("-l", "--lang", type=str, default='ch_sim')
        parser.add_argument("--backend", type=str, default='torch')
        parser.add_argument("--type", type=str, default='mobile')
        parser.add_argument("--output_format", type=str, default=None)
        parser.add_argument("--system_overrides", nargs='+', type=str, default=[])
        parser.add_argument("--det", type=str2bool, default=True)
        parser.add_argument("--rec", type=str2bool, default=True)
        parser.add_argument("--cls", type=str2bool, default=False)
        parser.add_argument("--verbose", type=str2bool, default=False)
        
        
        # params for detection engine
        parser.add_argument("--det_algorithm", type=str, default='CRAFT')
        parser.add_argument("--det_config_name", type=str, default='infer_det')
        parser.add_argument("--det_model_location", type=str, default=None)
        parser.add_argument("--det_batch_size", type=int, default=1) #batch_size is not yet implemented in the code
        parser.add_argument("--det_overrides", nargs='+', type=str, default=[])
        
        # params for recogniton engine
        parser.add_argument("--rec_algorithm", type=str, default='CRNN')
        parser.add_argument("--rec_config_name", type=str, default='infer_rec')
        parser.add_argument("--rec_model_location", type=str, default=None)
        parser.add_argument("--rec_batch_size", type=int, default=8) 
        parser.add_argument("--rec_overrides", nargs='+', type=str, default=[])
        parser.add_argument("--rec_whitelist", type=str, default=None)
        parser.add_argument("--rec_blacklist", type=str, default=None)
        
        # params for detection engine
        parser.add_argument("--cls_algorithm", type=str, default='CLS')
        parser.add_argument("--cls_config_name", type=str, default='infer_cls')
        parser.add_argument("--cls_model_location", type=str, default=None)
        parser.add_argument("--cls_batch_size", type=int, default=8)
        parser.add_argument("--cls_overrides", nargs='+', type=str, default=[])

        return parser.parse_args()
    else:
        return argparse.Namespace(
            hydra_conf_location=None,
            force_download=False,
            d = 'cpu',
            device='cpu',
            l = 'ch_sim',
            lang='ch_sim',
            type='mobile',
            backend='torch',
            output_format=None,
            system_overrides=[],
            verbose= False,
            
            det_algorithm='CRAFT',
            det_config_name='infer_det',
            det_model_location=None,
            det_batch_size=1,
            det_overrides=[],
            
            rec_algorithm='CRNN',
            rec_config_name='infer_rec',
            rec_model_location=None,
            rec_batch_size=8,
            rec_overrides=[],
            rec_whitelist=None,
            rec_blacklist=None,
            
            cls_algorithm='CLS',
            cls_config_name='infer_cls',
            cls_model_location=None,
            cls_batch_size=8,
            cls_overrides=[],
        )


class UCR(infer_system.TextSystem):
    def __init__(self, **kwargs):
        """
        ucr package
        args:
            **kwargs: other params show in ucr --help
        """
        args = parse_args(mMain=False, add_help=False)
        args.__dict__.update(**kwargs)
        
        force_download = args.force_download
        conf_location = args.hydra_conf_location
        det_algorithm = args.det_algorithm
        rec_algorithm = args.rec_algorithm
        cls_algorithm = args.cls_algorithm

        type = args.type
        backend = args.backend
        model_type = backend + '_' + type
        
        lang = args.l
        if args.lang!='ch_sim': 
            lang = args.lang
            if args.l!='ch_sim':
                print(f'WARNING: "l (={args.l})" argument is superseded by "lang (={args.lang})" argument! l=lang={lang}')

        device = args.d
        if args.device!='cpu':
            device = args.device
            if args.d!='cpu':
                print(f'WARNING: "d (={args.d})" argument is superseded by "device (={args.device})" argument! d=device={device}')

        if not conf_location:
            conf_location = maybe_download(os.path.join(BASE_DIR, '{}/conf'.format(VERSION)),model_urls['conf']['url'], force_download)
        
        with initialize_config_dir(config_dir=conf_location, job_name="infer_det"):
            model_type = 'torch_server' if det_algorithm == 'CRAFT' else model_type
            overrides = [f"{k}={v}" for k, v in model_urls[model_type]['det'][det_algorithm].items() if k!='url']
            overrides.extend(args.det_overrides)
            cfg = compose(config_name=args.det_config_name, overrides=overrides)    
            # print("Detection config:\n{}\n".format(cfg.pretty()))    
            
            config_det = OmegaConf.to_container(cfg)
            model_type = backend + '_' + type
            
        with initialize_config_dir(config_dir=conf_location, job_name="infer_rec"):
            model_type = 'torch_mobile' if lang != 'ch_sim' else model_type
            overrides = [f"{k}={v}" for k, v in model_urls[model_type]['rec'][lang].items() if k!='url']
            overrides.extend(args.rec_overrides)
            cfg = compose(config_name=args.rec_config_name, overrides=overrides)      
            # print("Recognition config:\n{}\n".format(cfg.pretty()))    
                  
            config_rec = OmegaConf.to_container(cfg)
            model_type = backend + '_' + type
            
        with initialize_config_dir(config_dir=conf_location, job_name="infer_cls"):
            overrides = [f"{k}={v}" for k, v in model_urls['torch_mobile']['cls'].items() if k!='url']
            overrides.extend(args.cls_overrides)
            cfg = compose(config_name=args.cls_config_name, overrides=overrides)    
            # print("Classification config:\n{}\n".format(cfg.pretty()))    
                    
            config_cls = OmegaConf.to_container(cfg)
        
        with initialize_config_dir(config_dir=conf_location, job_name="infer_system"):
            cfg = compose(config_name="infer_system", overrides=args.system_overrides)    
            # print("End-to-end config:\n{}\n".format(cfg.pretty()))    
                    
            config = OmegaConf.to_container(cfg)
            
        assert model_type in SUPPORT_MODEL_TYPE, 'TYPE_BACKEND must be of {} format, but got {}'.format(
                SUPPORT_MODEL_TYPE, model_type)
        
        config_det['device'] = config_rec['device'] = config_cls['device'] = device
        
        config_det['batch_size'] = args.det_batch_size
        config_rec['batch_size'] = args.rec_batch_size
        config_cls['batch_size'] = args.cls_batch_size
        
        config_rec['whitelist'] = args.rec_whitelist
        config_rec['blacklist'] = args.rec_blacklist
        config_rec['lang'] = lang
        
        # init model dir
        if not config_det['model_location']:
            model_type = 'torch_server' if det_algorithm == 'CRAFT' else model_type
            det_path = maybe_download(os.path.join(BASE_DIR, '{}/det'.format(VERSION)), model_urls[model_type]['det'][det_algorithm]['url'], force_download)
            config_det['model_location'] = os.path.join(det_path, 'model.pt')
            model_type = backend + '_' + type
        else:
            config_det['model_location'] = config_det['model_location'] # add abs path in parser args. 
        
        if not config_rec['model_location']:
            model_type = 'torch_mobile' if lang != 'ch_sim' else model_type
            rec_path = maybe_download(os.path.join(BASE_DIR, '{}/rec/{}'.format(VERSION, lang)), model_urls[model_type]['rec'][lang]['url'], force_download)
            config_rec['model_location'] = os.path.join(rec_path, 'model.pt')   
            
            __dir__ = os.path.dirname(__file__)
            config_rec['font_path'] = os.path.join(__dir__, config_rec['font_path'])  # TODO: Zip both font_path and char_dict_location in model_url
            config_rec['char_dict_location'] = os.path.join(__dir__, config_rec['char_dict_location'])   
        else:
            config_rec['model_location'] = config_rec['model_location'] # add abs path in parser args. 
            config_rec['font_path'] = config_rec['font_path'] # add abs path in .yaml file  
            config_rec['char_dict_location'] = config_rec['char_dict_location'] # add abs path in .yaml file       
            
        if not config_cls['model_location']:
            cls_path = maybe_download(os.path.join(BASE_DIR, '{}/cls'.format(VERSION)), model_urls['torch_mobile']['cls']['url'], force_download)
            config_cls['model_location'] = os.path.join(cls_path, 'model.pt')  
        else:
            config_cls['model_location'] = config_cls['model_location'] # add abs path in parser args.

        config['Detection'] = config_det
        config['Recognition'] = config_rec
        config['Classification'] = config_cls
        config["verbose"] = args.verbose
        config["output_format"] = args.output_format if args.output_format else config["output_format"]
        
        super().__init__(config)
    
    def predict(self, input=None, i=None, output=None, o=None, det=True, rec=True, cls=False):
        super().__call__(input, i, output, o, det, rec, cls)
        

def main():
    args = parse_args(mMain=True)
    
    ocr_engine = UCR(**(args.__dict__))
    result = ocr_engine.predict(input=args.input,
                            output=args.output,
                            det=args.det,
                            rec=args.rec,
                            cls=args.cls)
    if result is not None:
        for k,v in result.items():
            print(f'\n------------------------{k}:------------------------')
            print(tabulate(v, tablefmt="fancy_grid"))


model_urls = {
    
    'conf': {
        'url': 'https://docyard.s3.us-west-000.backblazeb2.com/UCR/conf/conf.zip'
    },
    
    'torch_server': {
        'det': {
            'DB': {
                'preprocess': 'det_db',
                'architecture': 'det_r50_vd_db',
                'postprocess': 'det_db',
                'url':
                'https://docyard.s3.us-west-000.backblazeb2.com/UCR/torch_server/det_ench_ppocr_server.zip'
            },
            'CRAFT': {
                'preprocess': 'det_craft',
                'architecture': 'det_vgg_craft',
                'postprocess': 'det_craft',
                'url':
                'https://docyard.s3.us-west-000.backblazeb2.com/UCR/torch_server/det_vgg_craft.zip'
            }    
        },
        'rec': {
            'ch_sim': {
                'preprocess': 'rec_ctc',
                'architecture': 'rec_ppocr_server',
                'postprocess': 'rec_ctc',
                'font_path': 'utils/fonts/simfang.ttf',
                'char_dict_location': 'utils/dict/ench_dict.txt',
                'url':
                'https://docyard.s3.us-west-000.backblazeb2.com/UCR/torch_server/rec_ench_ppocr_server.zip',                
            }
        }
    },

    'torch_mobile': {
        'det': {
            'DB': {
                'preprocess': 'det_db',
                'architecture': 'det_mv3_db',
                'postprocess': 'det_db',
                'url':
                'https://docyard.s3.us-west-000.backblazeb2.com/UCR/torch_mobile/det_ench_ppocr_mobile.zip',
            }  
        },
        'rec': {
            'ch_sim': {
                'preprocess': 'rec_ctc',
                'architecture': 'rec_ppocr_mobile',
                'postprocess': 'rec_ctc',
                'font_path': 'utils/fonts/simfang.ttf',
                'char_dict_location': 'utils/dict/ench_dict.txt',
                'url':
                'https://docyard.s3.us-west-000.backblazeb2.com/UCR/torch_mobile/rec_ench_ppocr_mobile.zip',                
            },
            'en': {
                'preprocess': 'rec_ctc',
                'architecture': 'rec_ppocr_mobile',
                'postprocess': 'rec_ctc',
                'font_path': 'utils/fonts/only_en.ttf',
                'char_dict_location': 'utils/dict/en_dict.txt',
                'url':
                'https://docyard.s3.us-west-000.backblazeb2.com/UCR/torch_mobile/rec_en_number_mobile.zip',                
            },
            'fr': {
                'preprocess': 'rec_ctc',
                'architecture': 'rec_ppocr_mobile',
                'postprocess': 'rec_ctc',
                'font_path': 'utils/fonts/french.ttf',
                'char_dict_location': 'utils/dict/french_dict.txt',
                'url':
                'https://docyard.s3.us-west-000.backblazeb2.com/UCR/torch_mobile/rec_french_mobile.zip',                
            },
            'de': {
                'preprocess': 'rec_ctc',
                'architecture': 'rec_ppocr_mobile',
                'postprocess': 'rec_ctc',
                'font_path': 'utils/fonts/german.ttf',
                'char_dict_location': 'utils/dict/german_dict.txt',
                'url':
                'https://docyard.s3.us-west-000.backblazeb2.com/UCR/torch_mobile/rec_german_mobile.zip',                
            },
            'ko': {
                'preprocess': 'rec_ctc',
                'architecture': 'rec_ppocr_mobile',
                'postprocess': 'rec_ctc',
                'font_path': 'utils/fonts/korean.ttf',
                'char_dict_location': 'utils/dict/korean_dict.txt',
                'url':
                'https://docyard.s3.us-west-000.backblazeb2.com/UCR/torch_mobile/rec_korean_mobile.zip',                
            },
            'ja': {
                'preprocess': 'rec_ctc',
                'architecture': 'rec_ppocr_mobile',
                'postprocess': 'rec_ctc',
                'font_path': 'utils/fonts/japan.ttc',
                'char_dict_location': 'utils/dict/japan_dict.txt',
                'url':
                'https://docyard.s3.us-west-000.backblazeb2.com/UCR/torch_mobile/rec_japan_mobile.zip',                
            },
            'ch_tra': {
                'preprocess': 'rec_ctc',
                'architecture': 'rec_ppocr_mobile',
                'postprocess': 'rec_ctc',
                'font_path': 'utils/fonts/chinese_cht.ttf',
                'char_dict_location': 'utils/dict/chinese_cht_dict.txt',
                'url':
                'https://docyard.s3.us-west-000.backblazeb2.com/UCR/torch_mobile/rec_chinese_cht_mobile.zip',                
            },
            'it': {
                'preprocess': 'rec_ctc',
                'architecture': 'rec_ppocr_mobile',
                'postprocess': 'rec_ctc',
                'font_path': 'utils/fonts/italian.ttf',
                'char_dict_location': 'utils/dict/it_dict.txt',
                'url':
                'https://docyard.s3.us-west-000.backblazeb2.com/UCR/torch_mobile/rec_it_mobile.zip',                
            },
            'es': {
                'preprocess': 'rec_ctc',
                'architecture': 'rec_ppocr_mobile',
                'postprocess': 'rec_ctc',
                'font_path': 'utils/fonts/xi.ttf',
                'char_dict_location': 'utils/dict/spanish_dict.txt',
                'url':
                'https://docyard.s3.us-west-000.backblazeb2.com/UCR/torch_mobile/rec_xi_mobile.zip',                
            },
            'pt': {
                'preprocess': 'rec_ctc',
                'architecture': 'rec_ppocr_mobile',
                'postprocess': 'rec_ctc',
                'font_path': 'utils/fonts/portugese.ttf',
                'char_dict_location': 'utils/dict/pu_dict.txt',
                'url':
                'https://docyard.s3.us-west-000.backblazeb2.com/UCR/torch_mobile/rec_pu_mobile.zip',                
            },
            'ru': {
                'preprocess': 'rec_ctc',
                'architecture': 'rec_ppocr_mobile',
                'postprocess': 'rec_ctc',
                'font_path': 'utils/fonts/russian.ttf',
                'char_dict_location': 'utils/dict/ru_dict.txt',
                'url':
                'https://docyard.s3.us-west-000.backblazeb2.com/UCR/torch_mobile/rec_ru_mobile.zip',                
            },
            'ar': {
                'preprocess': 'rec_ctc',
                'architecture': 'rec_ppocr_mobile',
                'postprocess': 'rec_ctc',
                'font_path': 'utils/fonts/arabic.ttf',
                'char_dict_location': 'utils/dict/ar_dict.txt',
                'url':
                'https://docyard.s3.us-west-000.backblazeb2.com/UCR/torch_mobile/rec_ar_mobile.zip',                
            },
            'hi': {
                'preprocess': 'rec_ctc',
                'architecture': 'rec_ppocr_mobile',
                'postprocess': 'rec_ctc',
                'font_path': 'utils/fonts/hindi.ttf',
                'char_dict_location': 'utils/dict/hi_dict.txt',
                'url':
                'https://docyard.s3.us-west-000.backblazeb2.com/UCR/torch_mobile/rec_hi_mobile.zip',                
            },
            'ug': {
                'preprocess': 'rec_ctc',
                'architecture': 'rec_ppocr_mobile',
                'postprocess': 'rec_ctc',
                'font_path': 'utils/fonts/uyghur.ttf',
                'char_dict_location': 'utils/dict/ug_dict.txt',
                'url':
                'https://docyard.s3.us-west-000.backblazeb2.com/UCR/torch_mobile/rec_ug_mobile.zip',                
            },
            'fa': {
                'preprocess': 'rec_ctc',
                'architecture': 'rec_ppocr_mobile',
                'postprocess': 'rec_ctc',
                'font_path': 'utils/fonts/persian.ttf',
                'char_dict_location': 'utils/dict/fa_dict.txt',
                'url':
                'https://docyard.s3.us-west-000.backblazeb2.com/UCR/torch_mobile/rec_fa_mobile.zip',                
            },
            'ur': {
                'preprocess': 'rec_ctc',
                'architecture': 'rec_ppocr_mobile',
                'postprocess': 'rec_ctc',
                'font_path': 'utils/fonts/urdu.ttf',
                'char_dict_location': 'utils/dict/ur_dict.txt',
                'url':
                'https://docyard.s3.us-west-000.backblazeb2.com/UCR/torch_mobile/rec_ur_mobile.zip',                
            },
            'rs_latin': {
                'preprocess': 'rec_ctc',
                'architecture': 'rec_ppocr_mobile',
                'postprocess': 'rec_ctc',
                'font_path': 'utils/fonts/serbian.ttf',
                'char_dict_location': 'utils/dict/rs_dict.txt',
                'url':
                'https://docyard.s3.us-west-000.backblazeb2.com/UCR/torch_mobile/rec_rs_mobile.zip',                
            },
            'oc': {
                'preprocess': 'rec_ctc',
                'architecture': 'rec_ppocr_mobile',
                'postprocess': 'rec_ctc',
                'font_path': 'utils/fonts/serbian.ttf',
                'char_dict_location': 'utils/dict/oc_dict.txt',
                'url':
                'https://docyard.s3.us-west-000.backblazeb2.com/UCR/torch_mobile/rec_oc_mobile.zip',                
            },
            'mr': {
                'preprocess': 'rec_ctc',
                'architecture': 'rec_ppocr_mobile',
                'postprocess': 'rec_ctc',
                'font_path': 'utils/fonts/marathi.ttf',
                'char_dict_location': 'utils/dict/mr_dict.txt',
                'url':
                'https://docyard.s3.us-west-000.backblazeb2.com/UCR/torch_mobile/rec_mr_mobile.zip',                
            },
            'ne': {
                'preprocess': 'rec_ctc',
                'architecture': 'rec_ppocr_mobile',
                'postprocess': 'rec_ctc',
                'font_path': 'utils/fonts/nepali.ttf',
                'char_dict_location': 'utils/dict/ne_dict.txt',
                'url':
                'https://docyard.s3.us-west-000.backblazeb2.com/UCR/torch_mobile/rec_ne_mobile.zip',                
            },
            'rs_cyrillic': {
                'preprocess': 'rec_ctc',
                'architecture': 'rec_ppocr_mobile',
                'postprocess': 'rec_ctc',
                'font_path': 'utils/fonts/serbian.ttf',
                'char_dict_location': 'utils/dict/rsc_dict.txt',
                'url':
                'https://docyard.s3.us-west-000.backblazeb2.com/UCR/torch_mobile/rec_rsc_mobile.zip',                
            },
            'bg': {
                'preprocess': 'rec_ctc',
                'architecture': 'rec_ppocr_mobile',
                'postprocess': 'rec_ctc',
                'font_path': 'utils/fonts/serbian.ttf',
                'char_dict_location': 'utils/dict/bg_dict.txt',
                'url':
                'https://docyard.s3.us-west-000.backblazeb2.com/UCR/torch_mobile/rec_bg_mobile.zip',                
            },
            'uk': {
                'preprocess': 'rec_ctc',
                'architecture': 'rec_ppocr_mobile',
                'postprocess': 'rec_ctc',
                'font_path': 'utils/fonts/serbian.ttf',
                'char_dict_location': 'utils/dict/uk_dict.txt',
                'url':
                'https://docyard.s3.us-west-000.backblazeb2.com/UCR/torch_mobile/rec_uk_mobile.zip',                
            },
            'be': {
                'preprocess': 'rec_ctc',
                'architecture': 'rec_ppocr_mobile',
                'postprocess': 'rec_ctc',
                'font_path': 'utils/fonts/serbian.ttf',
                'char_dict_location': 'utils/dict/be_dict.txt',
                'url':
                'https://docyard.s3.us-west-000.backblazeb2.com/UCR/torch_mobile/rec_be_mobile.zip',                
            },
            'te': {
                'preprocess': 'rec_ctc',
                'architecture': 'rec_ppocr_mobile',
                'postprocess': 'rec_ctc',
                'font_path': 'utils/fonts/telugu.ttf',
                'char_dict_location': 'utils/dict/te_dict.txt',
                'url':
                'https://docyard.s3.us-west-000.backblazeb2.com/UCR/torch_mobile/rec_te_mobile.zip',                
            },
            'kn': {
                'preprocess': 'rec_ctc',
                'architecture': 'rec_ppocr_mobile',
                'postprocess': 'rec_ctc',
                'font_path': 'utils/fonts/kannada.ttf',
                'char_dict_location': 'utils/dict/ka_dict.txt',
                'url':
                'https://docyard.s3.us-west-000.backblazeb2.com/UCR/torch_mobile/rec_ka_mobile.zip',                
            },
            'ta': {
                'preprocess': 'rec_ctc',
                'architecture': 'rec_ppocr_mobile',
                'postprocess': 'rec_ctc',
                'font_path': 'utils/fonts/tamil.ttf',
                'char_dict_location': 'utils/dict/ta_dict.txt',
                'url':
                'https://docyard.s3.us-west-000.backblazeb2.com/UCR/torch_mobile/rec_ta_mobile.zip',                
            }
        },
        'cls': {
            'preprocess': 'cls',
            'architecture': 'cls_ppocr_mobile',
            'postprocess': 'cls',
            'url':
            'https://docyard.s3.us-west-000.backblazeb2.com/UCR/torch_mobile/cls_ench_ppocr_mobile.zip'
        }
    }
}



# Copyright (c) 2020 PaddlePaddle Authors. All Rights Reserved.
#
# Modifications copyright (c) 2021 DocYard Authors. All Rights Reserve.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

from __future__ import absolute_import
from __future__ import division
from __future__ import print_function

import numpy as np
import hydra
from hydra.core.global_hydra import GlobalHydra
from omegaconf import OmegaConf
import os
import sys
import time
import logging
from tqdm import tqdm

log = logging.getLogger(__name__)
__dir__ = os.path.dirname(os.path.abspath(__file__))
sys.path.append(__dir__)
sys.path.append(os.path.abspath(os.path.join(__dir__, '../..')))

import cv2
import torch
import ucr.utils.annotation as utility
from ucr.utils.utility import get_image_file_list, check_and_read_gif
from ucr.core.preprocess import build_preprocess, preprocess
from ucr.core.postprocess import build_postprocess
from ucr.core.architecture import build_architecture


class TextDetector(object):
    def __init__(self, config):
        self.device = config['device']
        self.det_algorithm = config['Architecture']['algorithm']
        self.preprocess_op = build_preprocess(config['Preprocess'])
        self.postprocess_op = build_postprocess(config['Postprocess'])
        self.predictor = build_architecture(config['Architecture'])
        
        if self.device == 'cuda':
            self.device = torch.device('cuda') if torch.cuda.is_available() else torch.device('cpu')
        elif self.device == 'cpu':
            self.device = torch.device('cpu')
        else:
            log.error("wrong device selected! Choose eiter 'cuda' or 'cpu'")
            sys.exit(0)
            
        self.predictor.load_state_dict(torch.load(config['model_location'], map_location=self.device))
        self.predictor.eval()
        
    def order_points_clockwise(self, pts):
        """
        reference from: https://github.com/jrosebr1/imutils/blob/master/imutils/perspective.py
        # sort the points based on their x-coordinates
        """
        xSorted = pts[np.argsort(pts[:, 0]), :]

        # grab the left-most and right-most points from the sorted
        # x-roodinate points
        leftMost = xSorted[:2, :]
        rightMost = xSorted[2:, :]

        # now, sort the left-most coordinates according to their
        # y-coordinates so we can grab the top-left and bottom-left
        # points, respectively
        leftMost = leftMost[np.argsort(leftMost[:, 1]), :]
        (tl, bl) = leftMost

        rightMost = rightMost[np.argsort(rightMost[:, 1]), :]
        (tr, br) = rightMost

        rect = np.array([tl, tr, br, bl], dtype="float32")
        return rect

    def clip_det_res(self, points, img_height, img_width):
        for pno in range(points.shape[0]):
            points[pno, 0] = int(min(max(points[pno, 0], 0), img_width - 1))
            points[pno, 1] = int(min(max(points[pno, 1], 0), img_height - 1))
        return points

    def filter_tag_det_res(self, dt_boxes, image_shape):
        img_height, img_width = image_shape[0:2]
        dt_boxes_new = []
        for box in dt_boxes:
            box = self.order_points_clockwise(box)
            box = self.clip_det_res(box, img_height, img_width)
            rect_width = int(np.linalg.norm(box[0] - box[1]))
            rect_height = int(np.linalg.norm(box[0] - box[3]))
            if rect_width <= 3 or rect_height <= 3:
                continue
            dt_boxes_new.append(box)
        dt_boxes = np.array(dt_boxes_new)
        return dt_boxes

    def filter_tag_det_res_only_clip(self, dt_boxes, image_shape):
        img_height, img_width = image_shape[0:2]
        dt_boxes_new = []
        for box in dt_boxes:
            box = self.clip_det_res(box, img_height, img_width)
            dt_boxes_new.append(box)
        dt_boxes = np.array(dt_boxes_new)
        return dt_boxes

    def __call__(self, img):
        ori_im = img.copy()
        starttime = time.time()
        data = {'image': img}
        data = preprocess(data, self.preprocess_op)
        img, shape_list = data
        if img is None:
            return None, 0
        img = np.expand_dims(img, axis=0)
        shape_list = np.expand_dims(shape_list, axis=0)

        img = torch.as_tensor(img)
        input_tensors = img.to(self.device)
        self.predictor.to(self.device)
        
        output_tensors = self.predictor(input_tensors)
        outputs = []
        for output_tensor in output_tensors.values():
            output = output_tensor.cpu().data.numpy()
            outputs.append(output)
        
        preds = {}
        if self.det_algorithm == 'CRAFT':
            preds['text_map'] = outputs[0]
            preds['link_map'] = outputs[1]      

        elif self.det_algorithm == "EAST":
            preds['f_geo'] = outputs[0]
            preds['f_score'] = outputs[1]
        elif self.det_algorithm == 'SAST0' or self.det_algorithm == 'SAST1':
            preds['f_border'] = outputs[0]
            preds['f_score'] = outputs[1]
            preds['f_tco'] = outputs[2]
            preds['f_tvo'] = outputs[3]
        elif self.det_algorithm == 'DB':
            preds['maps'] = outputs[0]
        else:
            raise NotImplementedError

        post_result = self.postprocess_op(preds, shape_list)
        dt_boxes = post_result[0]['points']
        if self.det_algorithm == "SAST0": # TODO: Implement two versions of SAST both for det_sast_polygon True and False state
            dt_boxes = self.filter_tag_det_res_only_clip(dt_boxes, ori_im.shape)
        else:
            dt_boxes = self.filter_tag_det_res(dt_boxes, ori_im.shape)
                
        elapse = time.time() - starttime
        return dt_boxes, elapse


def main(cfg):    
    log.debug("Detection config:\n{}\n".format(cfg.pretty()))  
    config = OmegaConf.to_container(cfg)
    GlobalHydra.instance().clear()
    
    input = hydra.utils.to_absolute_path(config['input'])
    image_file_list = get_image_file_list(input)
    model_location = hydra.utils.to_absolute_path(config['model_location'])
    config['model_location'] = model_location
    text_detector = TextDetector(config)
    total_time = 0
    count = 0
    output = hydra.utils.to_absolute_path(config['output']) 
    if not os.path.exists(output):
        os.makedirs(output)
    for image_file in tqdm(image_file_list, colour='green', desc='Detection', unit = 'image'):
        count +=1
        img, flag = check_and_read_gif(image_file)
        if not flag:
            img = cv2.imread(image_file)
        if img is None:
            log.warning("error in loading image:{}".format(image_file))
            continue
        dt_boxes, elapse = text_detector(img)
        total_time += elapse
        src_im = utility.draw_text_det_res(dt_boxes, img)
        img_name_pure = os.path.split(image_file)[-1]
        img_path = os.path.join(output,
                                "det_{}".format(img_name_pure))
        cv2.imwrite(img_path, src_im)
        log.info("[{}/{}] Detection output is saved in --------- {}".format(count,len(image_file_list),img_path))
    log.info("\nTotal Prediction time for {} images:\t{:.5f} s".format(
        len(image_file_list), total_time))


if __name__ == "__main__":
    import argparse
    parser = argparse.ArgumentParser()
    parser.add_argument("--config_path", type=str, default='../../conf')
    parser.add_argument("--config_name", type=str, default='infer_det')
    
    args = parser.parse_args()
    
    main_wrapper = hydra.main(config_path=args.config_path, config_name=args.config_name)
    main_wrapper(main)()



# copyright (c) 2020 PaddlePaddle Authors. All Rights Reserve.
#
# Modifications copyright (c) 2021 DocYard Authors. All Rights Reserve.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#    http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import logging
import os
import imghdr
import cv2
import logging
import shutil
import requests
import sys
from tqdm import tqdm
import numpy as np

log = logging.getLogger(__name__)

def print_dict(d, logger, delimiter=0):
    """
    Recursively visualize a dict and
    indenting acrrording by the relationship of keys.
    """
    for k, v in sorted(d.items()):
        if isinstance(v, dict):
            log.info("{}{} : ".format(delimiter * " ", str(k)))
            print_dict(v, logger, delimiter + 4)
        elif isinstance(v, list) and len(v) >= 1 and isinstance(v[0], dict):
            log.info("{}{} : ".format(delimiter * " ", str(k)))
            for value in v:
                print_dict(value, logger, delimiter + 4)
        else:
            log.info("{}{} : {}".format(delimiter * " ", k, v))


def get_check_global_params(mode):
    check_params = ['device', 'max_text_length', 'image_shape', \
                    'image_shape', 'character_type', 'loss_type']
    if mode == "train_eval":
        check_params = check_params + [ \
            'train_batch_size_per_card', 'test_batch_size_per_card']
    elif mode == "test":
        check_params = check_params + ['test_batch_size_per_card']
    return check_params


def get_image_file_list(img_file):
    imgs_lists = []
    if img_file is None or not os.path.exists(img_file):
        raise Exception("not found any img file in {}".format(img_file))

    img_end = {'jpg', 'bmp', 'png', 'jpeg', 'rgb', 'tif', 'tiff', 'gif', 'GIF'}
    if os.path.isfile(img_file) and imghdr.what(img_file) in img_end:
        imgs_lists.append(img_file)
    elif os.path.isdir(img_file):
        for single_file in os.listdir(img_file):
            file_path = os.path.join(img_file, single_file)
            if os.path.isfile(file_path) and imghdr.what(file_path) in img_end:
                imgs_lists.append(file_path)
    if len(imgs_lists) == 0:
        raise Exception("not found any img file in {}".format(img_file))
    return imgs_lists


def check_and_read_gif(img_path):
    if os.path.basename(img_path)[-3:] in ['gif', 'GIF']:
        gif = cv2.VideoCapture(img_path)
        ret, frame = gif.read()
        if not ret:
            # logger = logging.getLogger('ucr')
            log.warning("Cannot read {}. This gif image maybe corrupted.")
            return None, False
        if len(frame.shape) == 2 or frame.shape[-1] == 1:
            frame = cv2.cvtColor(frame, cv2.COLOR_GRAY2RGB)
        imgvalue = frame[:, :, ::-1]
        return imgvalue, True
    return None, False


def download_with_progressbar(url, save_path):
    print("Downloading '{}' to '{}'".format(url, save_path))
    response = requests.get(url, stream=True)
    total_size_in_bytes = int(response.headers.get('content-length', 0))
    block_size = 1024  # 1 Kibibyte
    progress_bar = tqdm(total=total_size_in_bytes, unit='iB', unit_scale=True)
    with open(save_path, 'wb') as file:
        for data in response.iter_content(block_size):
            progress_bar.update(len(data))
            file.write(data)
    progress_bar.close()
    if total_size_in_bytes == 0 or progress_bar.n != total_size_in_bytes:
        print("Something went wrong while downloading model.")
        sys.exit(0)


def maybe_download(model_storage_directory, url, force_download=False):
    # using custom model 
    fname = url.split('/')[-1][0:-4]
    tmp_path = os.path.join(model_storage_directory, fname)
    if force_download or not os.path.exists(tmp_path):
        os.makedirs(tmp_path, exist_ok=True)
        download_with_progressbar(url, tmp_path+'.zip')
        shutil.unpack_archive(tmp_path+'.zip', tmp_path, 'zip')
        os.remove(tmp_path+'.zip')
    return tmp_path


def merge_text_boxes(dt_boxes, rec_res, **params):
    dt_boxes = np.asarray(dt_boxes)
    polys = np.empty((len(dt_boxes), 8))
    polys[:,0] = dt_boxes[:,0,0]
    polys[:,1] = dt_boxes[:,0,1]
    polys[:,2] = dt_boxes[:,1,0]
    polys[:,3] = dt_boxes[:,1,1]
    polys[:,4] = dt_boxes[:,2,0]
    polys[:,5] = dt_boxes[:,2,1]
    polys[:,6] = dt_boxes[:,3,0]
    polys[:,7] = dt_boxes[:,3,1]
    slope_ths = params["slope_thresh"]
    ycenter_ths = params["ycenter_thresh"]
    height_ths = params["height_thresh"]
    width_ths = params["width_thresh"]
    add_margin = params["add_margin"]

    horizontal_list, free_list_box, free_list_text, combined_list, merged_list_box, merged_list_text = [],[],[],[],[],[]

    for i, poly in enumerate(polys):
        slope_up = (poly[3]-poly[1])/np.maximum(10, (poly[2]-poly[0]))
        slope_down = (poly[5]-poly[7])/np.maximum(10, (poly[4]-poly[6]))
        if max(abs(slope_up), abs(slope_down)) < slope_ths:
            x_max = max([poly[0],poly[2],poly[4],poly[6]])
            x_min = min([poly[0],poly[2],poly[4],poly[6]])
            y_max = max([poly[1],poly[3],poly[5],poly[7]])
            y_min = min([poly[1],poly[3],poly[5],poly[7]])
            horizontal_list.append([x_min, x_max, y_min, y_max, 0.5*(y_min+y_max), y_max-y_min, rec_res[i][0], rec_res[i][1],str(poly)])
        else:
            height = np.linalg.norm( [poly[6]-poly[0],poly[7]-poly[1]])
            margin = int(1.44*add_margin*height)
            theta13 = abs(np.arctan( (poly[1]-poly[5])/np.maximum(10, (poly[0]-poly[4]))))
            theta24 = abs(np.arctan( (poly[3]-poly[7])/np.maximum(10, (poly[2]-poly[6]))))
            # do I need to clip minimum, maximum value here?
            x1 = poly[0] - np.cos(theta13)*margin
            y1 = poly[1] - np.sin(theta13)*margin
            x2 = poly[2] + np.cos(theta24)*margin
            y2 = poly[3] - np.sin(theta24)*margin
            x3 = poly[4] + np.cos(theta13)*margin
            y3 = poly[5] + np.sin(theta13)*margin
            x4 = poly[6] - np.cos(theta24)*margin
            y4 = poly[7] + np.sin(theta24)*margin

            free_list_box.append(np.array([[x1,y1],[x2,y2],[x3,y3],[x4,y4]]))
            free_list_text.append([rec_res[i][0], rec_res[i][1],str(poly), rec_res[i][0]])

    horizontal_list = sorted(horizontal_list, key=lambda item: item[4])

    # combine box
    new_box = []
    for poly in horizontal_list:

        if len(new_box) == 0:
            b_height = [poly[5]]
            b_ycenter = [poly[4]]
            new_box.append(poly)
        else:
            # comparable height and comparable y_center level up to ths*height
            if (abs(np.mean(b_height) - poly[5]) < height_ths*np.mean(b_height)) and (abs(np.mean(b_ycenter) - poly[4]) < ycenter_ths*np.mean(b_height)):
                b_height.append(poly[5])
                b_ycenter.append(poly[4])
                new_box.append(poly)
            else:
                b_height = [poly[5]]
                b_ycenter = [poly[4]]
                combined_list.append(new_box)
                new_box = [poly]
    combined_list.append(new_box)

    # merge list use sort again
    for boxes in combined_list:
        if len(boxes) == 1: # one box per line
            box = boxes[0]
            margin = int(add_margin*min(box[1]-box[0],box[5]))
            _x0 = _x3 = box[0]-margin
            _y0 = _y1 = box[2]-margin
            _x1 = _x2 = box[1]+margin
            _y2 = _y3 = box[3]+margin
            merged_list_box.append(np.array([[_x0,_y0],[_x1,_y1],[_x2,_y2],[_x3,_y3]]))
            merged_list_text.append([box[6], box[7], box[8], box[6]])
        else: # multiple boxes per line
            boxes = sorted(boxes, key=lambda item: item[0])

            merged_box, new_box = [],[]
            for box in boxes:
                if len(new_box) == 0:
                    b_height = [box[5]]
                    x_max = box[1]
                    new_box.append(box)
                else:
                    if abs(box[0]-x_max) < width_ths *(box[3]-box[2]): # merge boxes
                        x_max = box[1]
                        new_box.append(box)
                    else:
                        if (abs(np.mean(b_height) - box[5]) < height_ths*np.mean(b_height)) and (abs(box[0]-x_max) < width_ths *(box[3]-box[2])): # merge boxes
                            b_height.append(box[5])
                            x_max = box[1]
                            new_box.append(box)
                        else:
                            b_height = [box[5]]
                            x_max = box[1]
                            merged_box.append(new_box)
                            new_box = [box]
            if len(new_box) >0: merged_box.append(new_box)

            for mbox in merged_box:
                if len(mbox) != 1: # adjacent box in same line
                    # do I need to add margin here?
                    x_min = min(mbox, key=lambda x: x[0])[0]
                    x_max = max(mbox, key=lambda x: x[1])[1]
                    y_min = min(mbox, key=lambda x: x[2])[2]
                    y_max = max(mbox, key=lambda x: x[3])[3]
                    text_comb = str(mbox[0][6]) if isinstance(mbox[0][6], str) else ''
                    sum_score = mbox[0][7]
                    box_id = str(mbox[0][8])
                    text_id = str(mbox[0][6]) if isinstance(mbox[0][6], str) else ''
                    for val in range(len(mbox)-1):
                        if isinstance(mbox[val+1][6], str):
                            strin = mbox[val+1][6]
                        else:
                            strin = ''
                        text_comb += ' ' + strin
                        sum_score += mbox[val+1][7]
                        box_id += '|||' + str(mbox[val+1][8])
                        text_id += '|||' + strin 
                    avg_score = sum_score / len(mbox)
                    margin = int(add_margin*(y_max - y_min))

                    # merged_list.append([x_min-margin, x_max+margin, y_min-margin, y_max+margin, text_comb, avg_score])
                    _x0 = _x3 = x_min-margin
                    _y0 = _y1 = y_min-margin
                    _x1 = _x2 = x_max+margin
                    _y2 = _y3 = y_max+margin
                    merged_list_box.append(np.array([[_x0,_y0],[_x1,_y1],[_x2,_y2],[_x3,_y3]]))
                    merged_list_text.append([text_comb, avg_score, box_id, text_id])

                else: # non adjacent box in same line
                    box = mbox[0]

                    margin = int(add_margin*(box[3] - box[2]))
                    # merged_list.append([box[0]-margin,box[1]+margin,box[2]-margin,box[3]+margin, box[6], box[7]])
                    _x0 = _x3 = box[0]-margin
                    _y0 = _y1 = box[2]-margin
                    _x1 = _x2 = box[1]+margin
                    _y2 = _y3 = box[3]+margin
                    merged_list_box.append(np.array([[_x0,_y0],[_x1,_y1],[_x2,_y2],[_x3,_y3]]))
                    merged_list_text.append([box[6], box[7], box[8], box[6]])

    # may need to check if box is really in image
    return free_list_box, free_list_text, merged_list_box, merged_list_text

def sorted_boxes(dt_boxes):
    """
    Sort text boxes in order from top to bottom, left to right
    config:
        dt_boxes(array):detected text boxes with shape [4, 2]
    return:
        sorted boxes(array) with shape [4, 2]
    """
    num_boxes = dt_boxes.shape[0]
    sorted_boxes = sorted(dt_boxes, key=lambda x: (x[0][1], x[0][0]))
    _boxes = list(sorted_boxes)

    for i in range(num_boxes - 1):
        if abs(_boxes[i + 1][0][1] - _boxes[i][0][1]) < 10 and \
                (_boxes[i + 1][0][0] < _boxes[i][0][0]):
            tmp = _boxes[i]
            _boxes[i] = _boxes[i + 1]
            _boxes[i + 1] = tmp
    return _boxes